@@ -5,7 +5,7 @@ use strict;
 #use warnings;
 #no warnings 'uninitialized';
 package Readonly;
-our $VERSION = '2.01';
+our $VERSION = '2.02';
 $VERSION = eval $VERSION;
 
 # Autocroak (Thanks, MJD)
@@ -38,6 +38,72 @@ else {               # Modern perl doesn't need Readonly::XS
     $XSokay = 1;     # We're using the new built-ins so this is a white lie
 }
 
+# Undo setting readonly
+sub _SCALAR ($) {
+    my ($r_var) = @_;
+    if ($XSokay) {
+        Internals::SvREADONLY($r_var, 0) if is_sv_readonly($r_var);
+    }
+    else {
+        return if tied($r_var) !~ 'Readonly::Scalar';
+        my $r_scalar;
+        {
+            my $obj = tied $$r_var;
+            $r_scalar = $obj;
+        }
+        untie $r_var;
+        $r_var = $r_scalar;
+    }
+}
+
+sub _ARRAY (\@) {
+    my ($r_var) = @_;
+    return if tied(@$r_var) !~ 'Readonly::Array';
+    my $r_array;
+    {
+        my $obj = tied @$r_var;
+        $r_array = $obj;
+    }
+    untie @$r_var;
+    @$r_var = @$r_array;
+
+    # Recursively check child elements for references; clean if Readonly
+    foreach (@$r_var) {
+        my $_reftype = ref $_;
+        if ($_reftype eq 'SCALAR') { _SCALAR($_) }
+        elsif ($_reftype eq 'ARRAY') {
+            _ARRAY($_);
+        }
+        elsif ($_reftype eq 'HASH') {
+            _HASH($_);
+        }
+    }
+}
+
+sub _HASH (\%) {
+    my ($r_var) = @_;
+    return if tied(%$r_var) !~ 'Readonly::Hash';
+    my $r_hash;
+    {
+        my $obj = tied %$r_var;
+        $r_hash = $obj;
+    }
+    untie %$r_var;
+    %$r_var = %$r_hash;
+
+    # Recursively check child elements for references; clean if Readonly
+    foreach (values %$r_var) {
+        my $_reftype = ref $_;
+        if ($_reftype eq 'SCALAR') { _SCALAR($_) }
+        elsif ($_reftype eq 'ARRAY') {
+            _ARRAY(@$_);
+        }
+        elsif ($_reftype eq 'HASH') {
+            _HASH(%$_);
+        }
+    }
+}
+
 # Common error messages, or portions thereof
 use vars qw/$MODIFY $REASSIGN $ODDHASH/;
 $MODIFY   = 'Modification of a read-only value attempted';
@@ -49,6 +115,11 @@ $ODDHASH  = 'May not store an odd number of values in a hash';
 # ----------------
 package Readonly::Scalar;
 
+sub STORABLE_freeze {
+    my ($self, $cloning) = @_;
+    Readonly::_SCALAR($$self) if $cloning;
+}
+
 sub TIESCALAR {
     my $whence
         = (caller 2)[3];    # Check if naughty user is trying to tie directly.
@@ -66,13 +137,19 @@ sub FETCH {
     my $self = shift;
     return $$self;
 }
-*STORE = *UNTIE = sub { Readonly::croak $Readonly::MODIFY};
+*STORE = sub { Readonly::croak $Readonly::MODIFY };
+*UNTIE = sub { Readonly::croak $Readonly::MODIFY if caller() ne 'Readonly' };
 
 # ----------------
 # Read-only arrays
 # ----------------
 package Readonly::Array;
 
+sub STORABLE_freeze {
+    my ($self, $cloning) = @_;
+    Readonly::_ARRAY(@$self) if $cloning;
+}
+
 sub TIEARRAY {
     my $whence
         = (caller 1)[3];    # Check if naughty user is trying to tie directly.
@@ -104,13 +181,19 @@ BEGIN {
     } if $] >= 5.006;    # couldn't do "exists" on arrays before then
 }
 *STORE = *STORESIZE = *EXTEND = *PUSH = *POP = *UNSHIFT = *SHIFT = *SPLICE
-    = *CLEAR = *UNTIE = sub { Readonly::croak $Readonly::MODIFY};
+    = *CLEAR = sub { Readonly::croak $Readonly::MODIFY};
+*UNTIE = sub { Readonly::croak $Readonly::MODIFY if caller() ne 'Readonly' };
 
 # ----------------
 # Read-only hashes
 # ----------------
 package Readonly::Hash;
 
+sub STORABLE_freeze {
+    my ($self, $cloning) = @_;
+    Readonly::_HASH(%$self) if $cloning;
+}
+
 sub TIEHASH {
     my $whence
         = (caller 1)[3];    # Check if naughty user is trying to tie directly.
@@ -145,7 +228,8 @@ sub NEXTKEY {
     my $self = shift;
     return scalar each %$self;
 }
-*STORE = *DELETE = *CLEAR = *UNTIE = sub { Readonly::croak $Readonly::MODIFY};
+*STORE = *DELETE = *CLEAR = sub { Readonly::croak $Readonly::MODIFY};
+*UNTIE = sub { Readonly::croak $Readonly::MODIFY if caller() ne 'Readonly'; };
 
 # ----------------------------------------------------------------
 # Main package, containing convenience functions (so callers won't
@@ -289,6 +373,23 @@ sub Hash (\%;@) {
     return tie %$href, 'Readonly::Hash', @values;
 }
 
+sub Clone(\[$@%]) {
+    require Storable;
+    my $reftype = ref $_[0];
+    my $retval  = Storable::dclone($_[0]);
+    if ($reftype eq 'SCALAR') {
+        _SCALAR($retval);
+        return $$retval;
+    }
+    elsif ($reftype eq 'ARRAY') {
+        _ARRAY(@$retval);
+    }
+    elsif ($reftype eq 'HASH') {
+        _HASH(%$retval);
+    }
+    return $retval;
+}
+
 # Common entry-point for all supported data types
 eval q{sub Readonly} . ($] < 5.008 ? '' : '(\[$@%]@)') . <<'SUB_READONLY';
 {
@@ -616,6 +717,17 @@ ones. For example:
 
 =back
 
+=head1 Cloning
+
+When cloning using L<Storable> or L<Clone> you will notice that the value stays
+readonly, which is correct. If you want to clone the value without copying 
+readonly flag, use the C<Clone> function:
+
+    Readonly::Scalar my $scalar = 'string';
+    my $scalar_clone = Readonly::Clone $scalar_clone;
+
+    $scalar_clone .= 'foo'; # no error
+
 =head1 Examples
 
 These are a few very simple examples:
@@ -697,7 +809,7 @@ const-related modules that have cropped up since Readonly was released in
 To 'fix' this, Readonly::XS was written. If installed, Readonly::XS used the
 internal methods C<SvREADONLY> and C<SvREADONLY_on> to lock simple scalars. On
 the surface, everything was peachy but things weren't the same behind the
-scenes. In edge cases, code perfromed very differently if Readonly::XS was
+scenes. In edge cases, code performed very differently if Readonly::XS was
 installed and because it wasn't a required dependency in most code, it made
 downstream bugs very hard to track.
 
@@ -734,9 +846,6 @@ If email is better for you, L<my address is mentioned below|/"Author"> but I
 would rather have bugs sent through the issue tracker found at
 http://github.com/sanko/readonly/issues.
 
-Please check the TODO file included with this distribution in case your bug
-is already known (...I probably won't file bug reports to myself).
-
 =head1 Acknowledgements
 
 Thanks to Slaven Rezic for the idea of one common function (Readonly) for all
@@ -758,7 +867,7 @@ Original author: Eric J. Roode, roode@cpan.org
 
 =head1 License and Legal
 
-Copyright (C) 2013, 2014 by Sanko Robinson <sanko@cpan.org>
+Copyright (C) 2013-2016 by Sanko Robinson <sanko@cpan.org>
 
 Copyright (c) 2001-2004 by Eric J. Roode. All Rights Reserved.
 
